import secrets


def strtobool(val: str) -> bool:
    """
    Convert a string representation of truth to True or False.

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True

    if val in ("n", "no", "f", "false", "off", "0"):
        return False

    msg = f"invalid truth value {val}"
    raise ValueError(msg)


def generate_canary_word(prefix: str = "BANKS[", suffix: str = "]", token_length: int = 8) -> str:
    return f"{prefix}{secrets.token_hex(token_length // 2)}{suffix}"



import pytest
import regex as re

from banks.utils import generate_canary_word, strtobool


def test_generate_canary_word_defaults():
    default = generate_canary_word()
    assert re.match(r"BANKS\[.{8}\]", default)


def test_generate_canary_word_params():
    only_token = generate_canary_word(prefix="", suffix="", token_length=16)
    assert re.match(r".{16}", only_token)

    only_prefix = generate_canary_word(prefix="foo", suffix="")
    assert re.match(r"foo.{8}", only_prefix)

    only_suffix = generate_canary_word(prefix="", suffix="foo")
    assert re.match(r".{8}foo", only_suffix)


def test_strtobool_error():
    with pytest.raises(ValueError):
        strtobool("42")


@pytest.mark.parametrize(
    "test_input,expected",
    [
        ("y", True),
        ("yes", True),
        ("t", True),
        ("true", True),
        ("on", True),
        ("1", True),
        ("n", False),
        ("no", False),
        ("f", False),
        ("false", False),
        ("off", False),
        ("0", False),
        pytest.param("42", True, marks=pytest.mark.xfail),
    ],
)
def test_strtobool(test_input, expected):
    assert strtobool(test_input) == expected



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from banks.extensions.generate import GenerateExtension
from banks.extensions.inference_endpoint import HFInferenceEndpointsExtension

__all__ = ("GenerateExtension", "HFInferenceEndpointsExtension")



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
import html
import os

import requests
from jinja2 import nodes
from jinja2.ext import Extension


class HFInferenceEndpointsExtension(Extension):
    """
    `inference_endpoint` can be used to call the Hugging Face Inference Endpoint API
    passing a prompt to get back some content.

    Example:
        ```jinja
        {% inference_endpoint "write a tweet with positive sentiment", "https://foo.aws.endpoints.huggingface.cloud" %}
        Life is beautiful, full of opportunities & positivity
        ```
    """

    # a set of names that trigger the extension.
    tags = {"inference_endpoint"}  # noqa

    def parse(self, parser):
        # We get the line number of the first token so that we can give
        # that line number to the nodes we create by hand.
        lineno = next(parser.stream).lineno

        # The args passed to the extension:
        # - the prompt text used to generate new text
        args = [parser.parse_expression()]
        # - second param after the comma,  the inference endpoint URL
        parser.stream.skip_if("comma")
        args.append(parser.parse_expression())

        return nodes.Output([self.call_method("_call_endpoint", args)]).set_lineno(lineno)

    def _call_endpoint(self, text, endpoint):
        """
        Helper callback.
        """
        access_token = os.environ.get("HF_ACCESS_TOKEN")
        response = requests.post(
            endpoint, json={"inputs": text}, headers={"Authorization": f"Bearer {access_token}"}, timeout=30
        )
        response_body = response.json()

        if response_body:
            return html.unescape(response_body[0].get("generated_text", ""))
        return ""



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from typing import cast

from jinja2 import nodes
from jinja2.ext import Extension
from litellm import acompletion, completion
from litellm.types.utils import ModelResponse

from banks.errors import CanaryWordError
from banks.prompt import Prompt

DEFAULT_MODEL = "gpt-3.5-turbo"
SYSTEM_PROMPT = Prompt("{{canary_word}} You are a helpful assistant.")


# This function exists for documentation purpose.
def generate(model_name: str):  # pylint: disable=W0613
    """
    `generate` can be used to call the LiteLLM API passing the tag text as a prompt and get back some content.

    Example:
        ```jinja
        {% generate "write a tweet with positive sentiment" "gpt-3.5-turbo" %}
        Feeling grateful for all the opportunities that come my way! #positivity #productivity
        ```
    """


class GenerateExtension(Extension):
    # a set of names that trigger the extension.
    tags = {"generate"}  # noqa

    def parse(self, parser):
        # We get the line number of the first token so that we can give
        # that line number to the nodes we create by hand.
        lineno = next(parser.stream).lineno

        # The args passed to the extension:
        # - the prompt text used to generate new text
        # - (optional) the name of the model use to generate new text
        args = [parser.parse_expression()]

        # If there is a comma, the user provided the model name. If not, use
        # None as the second parameter.
        if parser.stream.skip_if("comma"):
            args.append(parser.parse_expression())
        else:
            args.append(nodes.Const(None))

        if parser.environment.is_async:
            return nodes.Output([self.call_method("_agenerate", args)]).set_lineno(lineno)
        return nodes.Output([self.call_method("_generate", args)]).set_lineno(lineno)

    def _generate(self, text, model_name=DEFAULT_MODEL):
        """
        Helper callback.

        To tweak the prompt used to generate content, change the variable `messages` .
        """
        messages = [
            {"role": "system", "content": SYSTEM_PROMPT.text()},
            {"role": "user", "content": text},
        ]
        response: ModelResponse = cast(ModelResponse, completion(model=model_name, messages=messages))
        return self._get_content(response)

    async def _agenerate(self, text, model_name=DEFAULT_MODEL):
        """
        Helper callback.

        To tweak the prompt used to generate content, change the variable `messages` .
        """
        messages = [
            {"role": "system", "content": SYSTEM_PROMPT.text()},
            {"role": "user", "content": text},
        ]
        response: ModelResponse = cast(ModelResponse, await acompletion(model=model_name, messages=messages))
        return self._get_content(response)

    def _get_content(self, response: ModelResponse) -> str:
        content = response["choices"][0]["message"]["content"]
        if SYSTEM_PROMPT.canary_leaked(content):
            msg = "The system prompt has leaked into the response, possible prompt injection!"
            raise CanaryWordError(msg)
        return content



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from enum import Enum

from pydantic import BaseModel

# pylint: disable=invalid-name


class ContentBlockType(str, Enum):
    text = "text"
    image = "image"


class MediaTypeBlockType(str, Enum):
    image_jpeg = "image/jpeg"
    image_png = "image/png"
    image_gif = "image/gif"
    image_webp = "image/webp"


class CacheControl(BaseModel):
    type: str = "ephemeral"


class Source(BaseModel):
    type: str = "base64"
    media_type: MediaTypeBlockType
    data: str

    class Config:
        use_enum_values = True


class ContentBlock(BaseModel):
    type: ContentBlockType
    cache_control: CacheControl | None = None
    text: str | None = None
    source: Source | None = None

    class Config:
        use_enum_values = True


ChatMessageContent = list[ContentBlock] | str


class ChatMessage(BaseModel):
    role: str
    content: str | ChatMessageContent



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from jinja2 import PackageLoader

from banks import env


def test_default_loader():
    assert type(env.loader) is PackageLoader



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from .cache_control import cache_control
from .lemmatize import lemmatize

__all__ = (
    "cache_control",
    "lemmatize",
)
